// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_TECTONIC_LOCALBISECTIONSOLVER_HH
#define DUNE_TECTONIC_LOCALBISECTIONSOLVER_HH

#include <dune/tnnmg/problem-classes/bisection.hh>

/**
 * \brief A local solver for quadratic obstacle problems with nonlinearity
 * using bisection
 */
class LocalBisectionSolver
{
public:
  template<class Vector, class Functional, class BitVector>
  void operator()(Vector& x, const Functional& f, const BitVector& ignore) const {
      if (ignore.all()) {
          return;
      }

      double safety = 1e-14;
      if (f.defectUpper() - f.defectLower() < safety) {
          x = f.origin();
          return;
      }

      int bisectionsteps = 0;
      const Bisection globalBisection;

      if (ignore.any()) {
            auto&& restrictedf = f.getIgnoreFunctional(ignore);
            const auto alpha = globalBisection.minimize(restrictedf, 0.0, 0.0, bisectionsteps);

#ifdef NEW_TNNMG_COMPUTE_ITERATES_DIRECTLY
            x = restrictedf.origin();
            auto correction = restrictedf.direction();
            correction *= alpha;
            x += correction;
#else
            auto x = restrictedf.direction();
            x *= alpha;
#endif


      } else { // ignore.none()
          const auto alpha = globalBisection.minimize(f, 0.0, 0.0, bisectionsteps);

#ifdef NEW_TNNMG_COMPUTE_ITERATES_DIRECTLY
          x = f.origin();
          auto correction = f.direction();
          correction *= alpha;
          x += correction;
#else
          auto x = f.direction();
          x *= alpha;
#endif

      }
  }
};

#endif




